--[[
# Copyright 2001-2014 Cisco Systems, Inc. and/or its affiliates. All rights
# reserved.
#
# This file contains proprietary Detector Content created by Cisco Systems,
# Inc. or its affiliates ("Cisco") and is distributed under the GNU General
# Public License, v2 (the "GPL").  This file may also include Detector Content
# contributed by third parties. Third party contributors are identified in the
# "authors" file.  The Detector Content created by Cisco is owned by, and
# remains the property of, Cisco.  Detector Content from third party
# contributors is owned by, and remains the property of, such third parties and
# is distributed under the GPL.  The term "Detector Content" means specifically
# formulated patterns and logic to identify applications based on network
# traffic characteristics, comprised of instructions in source code or object
# code form (including the structure, sequence, organization, and syntax
# thereof), and all documentation related thereto that have been officially
# approved by Cisco.  Modifications are considered part of the Detector
# Content.
--]]
--[[
detection_name: Kerberos
version: 4
description: Kerberos
--]]

require "DetectorCommon"

--require('debugger')

--local DC = require("DetectorCommon")
local DC = DetectorCommon

DetectorPackageInfo = {
    name =  "Kerberos",
    proto =  DC.ipproto.udp,
    server = {
        init =  'DetectorInit',
        validate =  'DetectorValidator',
    }
}

gServiceId = 20011
gServiceName = 'Kerberos'

--port based detection - needed when not using CSD tables
gPorts = {
    {DC.ipproto.udp, 88},
}

gAppRegistry = {
	--AppIdValue          Extracts Info
	---------------------------------------
	{701,		         1}
}

function serviceInProcess(context)

    local flowFlag = context.detectorFlow:getFlowFlag(DC.flowFlags.serviceDetected)

    if ((not flowFlag) or (flowFlag == 0)) then
        gDetector:inProcessService()
    end

    DC.printf('%s: Inprocess, packetCount: %d\n', gServiceName, context.packetCount);
    return DC.serviceStatus.inProcess
end

function serviceSuccess(context)
    local flowFlag = context.detectorFlow:getFlowFlag(DC.flowFlags.serviceDetected)

    if ((not flowFlag) or (flowFlag == 0)) then
        gDetector:addService(gServiceId, "", context.version, 701)
    end

    DC.printf('%s: Detected, packetCount: %d\n', gServiceName, context.packetCount);
    return DC.serviceStatus.success
end

function serviceFail(context)
    local flowFlag = context.detectorFlow:getFlowFlag(DC.flowFlags.serviceDetected)

    if ((not flowFlag) or (flowFlag == 0)) then
        gDetector:failService()
    end

    context.detectorFlow:clearFlowFlag(DC.flowFlags.continue)
    DC.printf('%s: Failed, packetCount: %d\n', gServiceName, context.packetCount);
    return DC.serviceStatus.nomatch
end

function registerPortsPatterns()

    --register port based detection
    for i,v in ipairs(gPorts) do
        gDetector:addPort(v[1], v[2])
    end


	for i,v in ipairs(gAppRegistry) do
		pcall(function () gDetector:registerAppId(v[1],v[2]) end)
	end

end


--[[ Core engine calls DetectorInit() to initialize a detector.
--]]
function DetectorInit( detectorInstance)

    --print (gServiceName .. ': DetectorInit()')

    gDetector = detectorInstance
    gDetector:init(gServiceName, 'DetectorValidator', 'DetectorFini')
    registerPortsPatterns()

    return gDetector
end

function getStringPlain (data, length)
    local stringValue=""
    local index=0
    while (index <  length) do
        stringValue = string.format('%s%.2x', stringValue, string.byte(data,index+1))
        index = index + 1
    end
    return stringValue
end


--[[Validator function registered in DetectorInit()

    (1+dir) and (2-dir) logic takes care of symmetric request response case. Once connection is established,
    client (server) can send request and server (client) should send a response.
--]]
function DetectorValidator()
    local context = {}
    context.detectorFlow = gDetector:getFlow()
    context.packetDataLen = gDetector:getPacketSize()
    context.packetDir = gDetector:getPacketDir()
    context.srcIp = gDetector:getPktSrcAddr()
    context.dstIp = gDetector:getPktDstAddr()
    context.srcPort = gDetector:getPktSrcPort()
    context.dstPort = gDetector:getPktDstPort()
    context.flowKey = context.detectorFlow:getFlowKey()
    context.packetCount = gDetector:getPktCount()
    local size = context.packetDataLen
    local dir = context.packetDir
    local srcPort = context.srcPort
    local dstPort = context.dstPort
    local flowKey = context.flowKey
    local current_index = 0


    DC.printf('%s:DetectorValidator(): packetCount %d, dir %d, size %d, srcPort %d\n', gServiceName, context.packetCount, dir, size, srcPort)
    if (size == 0 or dir == 0) then
        return serviceInProcess(context)
    end

    if (dir == DC.flowDirection.fromResponder and size >= 12 and srcPort == 88) then 
            matched, msg_type_raw, pattern_one_raw = gDetector:getPcreGroups('(.)(..)', current_index)
            msg_type = getStringPlain(msg_type_raw, 1)
            pattern_one = DC.getShortHostFormat(pattern_one_raw)
            if (msg_type ~= "6b" and msg_type ~= "6d" 
                and msg_type ~= "7e" and msg_type ~= "7d") then 
                --  there may have been a four-byte "record marker" 
                current_index = current_index + 4
                matched, msg_type_raw, pattern_one_raw = gDetector:getPcreGroups('(.)(..)', current_index)
                msg_type = getStringPlain(msg_type_raw, 2)
                pattern_one = DC.getShortHostFormat(pattern_one_raw)
                if (msg_type ~= "6b" and msg_type ~= "6d" 
                    and msg_type ~= "7e" and msg_type ~= "7d") then
                    return serviceFail(context)
                end
            end
            -- 7d and 7e are kerberos error messages whose morphology is different than 6b and 6d packets
            if (msg_type == "7e" or msg_type == "7d") then
                matched, version = gDetector:getPcreGroups('\160\003\002\001(.)', current_index)
                if matched then
                    context.version = getStringPlain(version, 1)
                    return serviceSuccess(context)
                else
                    return serviceFail(context)
                end
            end
            current_index = current_index + 5
            matched, pattern_two_raw = gDetector:getPcreGroups('(..)', current_index)
            pattern_two = DC.getShortHostFormat(pattern_two_raw)
            if (pattern_one ~= pattern_two) then
                return serviceFail(context)
            end
            current_index = current_index + 3
            matched, pad_sequence_raw = gDetector:getPcreGroups('(....)', current_index)
            pad_sequence = getStringPlain(pad_sequence_raw, 4)
            if (pad_sequence ~= "a0030201") then
                return serviceFail(context)
            end
            current_index = current_index + 4
            matched, version = gDetector:getPcreGroups('(.)', current_index)
            context.version = getStringPlain(version, 1)
            return serviceSuccess(context)
       
    end

    return serviceFail(context)
    --return serviceFail(context)
end

--[[Required DetectorFini function
--]]
function DetectorFini()
    --print (gServiceName .. ': DetectorFini()')
end
